#include <iostream>
#include <fstream>
#include <cmath>

#ifdef WIN32

#include <cv.h>
#include <highgui.h>
#include <blobresult.h>

#pragma comment(lib, "cv.lib" )
#pragma comment(lib, "cxcore.lib")
#pragma comment(lib, "highgui.lib")
#pragma comment(lib, "cvblobslib.lib")

#else

#include <opencv/cv.h>
#include <opencv/highgui.h>

#endif

using namespace std;

#define FEATURE_COUNT		100
#define QUALITY				0.01
#define MIN_DISTANCE		10

CvScalar hsv2rgb( float hue );

IplImage *pImgOriginal;
IplImage *pImgHsv;
IplImage *pHueChannel;
IplImage *pMask;
IplImage *pMask2 = NULL;
IplImage *pBackProject = NULL;
IplImage *pBackImage = NULL;
IplImage *pDebugImage = NULL;

IplImage *pImgHist = NULL;

CvHistogram *pHist = NULL;
CvScalar maxColor;

int saturationMax = 255,
	saturationMin = 0,
	hueValue = 0,
	hueValue2 = 0;

int nFrames = 0;

int hueDim = 32;
float hueRange[] = { 0, 180 };
float *pHueRange = hueRange;
float maxValue = 0;
int maxIndex = 0;
CvRect search_area;
bool bFirstRun = true;

// VARIABILI PER IL FLOW FEATURES
IplImage *pGrayF = NULL,
		 *pPrevGrayF = NULL,
		 *pPyramidF = NULL,
		 *pPrevPyramidF = NULL,
		 *pEigF = NULL,
		 *pTempF = NULL,
		 *pSwapTempF = NULL;

char *statusF = NULL;
bool bInitFeatures = true;
int countF = FEATURE_COUNT;
int flagsF = 0;
int feat_count = FEATURE_COUNT;
CvRect featRect;
CvPoint featCenter;
CvBox2D feat_box;

CvPoint2D32f *pPointsF[2] = { 0 },
			 *pSwapPointsF = NULL;

/********************** VARIABILI PER IL TEMPLATE MATCHING **********************/
IplImage *pImageGrayT = NULL,
		 *pTemplateT = NULL,
		 *pMatchResultT = NULL;

bool bInitTemplate = true;
/********************************************************************************/

// VARIABILI PER LA SELZIONE DELL'OGGETTO DA SEGUIRE
CvRect sel;
CvPoint origin;
bool bSel = false;
bool bRun = false;

void BlobDetector( string file );
void FlowFeatures( IplImage *pFrame );
void TemplateMatch( IplImage *pFrame );

int main( int argc, char **argv ){
	if ( argc < 2 ){
		cout << "HELP: cv filename(*.avi)" << endl;
		return 1;
	}

	BlobDetector( argv[1] );
}

void on_track( int dummy ){

}

void on_mouse( int event, int x, int y, int flags, void *param ){
	if ( !pImgOriginal )
		return;

	if ( pImgOriginal->origin )
		y = pImgOriginal->height - y;


	if ( bSel ){
		sel.x = MIN( x, origin.x );
		sel.y = MIN( y, origin.y );
		sel.width = sel.x + CV_IABS( x - origin.x );
		sel.height = sel.y + CV_IABS( y - origin.y );

		sel.x = MAX( sel.x, 0 );
        sel.y = MAX( sel.y, 0 );
        sel.width = MIN( sel.width, pImgOriginal->width );
        sel.height = MIN( sel.height, pImgOriginal->height );
        sel.width -= sel.x;
        sel.height -= sel.y;
	}

	switch ( event ){
		case CV_EVENT_LBUTTONDOWN:
			origin = cvPoint(x,y);
			sel = cvRect(x,y,0,0);
			bSel = true;
			bRun = false;
			bFirstRun = true;
			break;
		case CV_EVENT_LBUTTONUP:
			bSel = false;
			bRun = true;
			break;
	}
}

void BlobDetector( string file ){
	cout << "Blob tracking: " << file.c_str() << endl;

	featRect = cvRect(0,0,0,0);
	featCenter = cvPoint(0,0);

	CvCapture *cam = ::cvCaptureFromAVI( file.c_str() );
	CvConnectedComp track_comp;
	CvBox2D track_box;

	IplImage *featMask = NULL;


	cvNamedWindow( "original", CV_WINDOW_AUTOSIZE );	cvMoveWindow( "original", 10, 10 );
	cvNamedWindow( "backproject", CV_WINDOW_AUTOSIZE );	cvMoveWindow( "backproject", 360, 10 );
	cvNamedWindow( "hist", CV_WINDOW_AUTOSIZE );		cvMoveWindow( "hist", 10, 400 );

	cvCreateTrackbar( "Hue1", "hist", &hueValue, 180, on_track );
	cvCreateTrackbar( "sat min", "hist", &saturationMin, 255, on_track );
	cvCreateTrackbar( "sat max", "hist", &saturationMax, 255, on_track );

	cvSetTrackbarPos( "sat min", "hist", 30 );
	cvSetMouseCallback( "original", on_mouse, NULL );


	
//	IplImage *frame = cvLoadImage( "4.jpg" );
	do  {
		IplImage *frame = cvQueryFrame(cam);

		// SENZA DI QUESTO IL VIDEO E' TROPPO VELOCE ( con la webcam il problema non si presenta )
#ifdef WIN32
		::Sleep(20);
#else
		usleep(20000);
#endif

		if ( !frame )
			break;
		
		if ( !pImgOriginal ){
			
			pImgOriginal = cvCreateImage( cvGetSize(frame), 8, 3 );						//Immagine catturata via cam
			pImgHsv = cvCreateImage( cvGetSize(pImgOriginal), IPL_DEPTH_8U, 3 );		//Immagine convertita in hsv
			pHueChannel = cvCreateImage( cvGetSize(pImgOriginal), IPL_DEPTH_8U, 1 );	//Canale hue
			pMask = cvCreateImage( cvGetSize(pImgOriginal), IPL_DEPTH_8U, 1 );			//Maskera sogliatura colore
			pBackProject = cvCreateImage( cvGetSize(pImgOriginal), IPL_DEPTH_8U, 1 );	//backproject
			pBackImage = cvCreateImage( cvGetSize( pImgOriginal ), 8, 3 );				//backproject image - utility
			pDebugImage = cvCreateImage( cvGetSize( pImgOriginal ), 8, 3 );				// DEBUG

			pHist = cvCreateHist( 1, &hueDim, CV_HIST_ARRAY, &pHueRange, 1 );			//Istogramma colore
			pImgHist = cvCreateImage( cvSize( 320, 100 ), IPL_DEPTH_8U, 3 );		//Immaggine istogramma - utility

			search_area = cvRect( 0, 0, pImgOriginal->width, pImgOriginal->height );

			pImgOriginal->origin = frame->origin;

			featMask = cvCreateImage( cvGetSize(pImgOriginal), 8, 1 );
		}
		
		cvCopy( frame, pImgOriginal );

		if ( pImgOriginal == NULL )
			return;

		cvCopy( pImgOriginal, pDebugImage );

		if ( bSel && sel.width > 0  && sel.height > 0 ){
			cvSetImageROI( pDebugImage, sel );
			cvXorS( pDebugImage, cvScalarAll(255), pDebugImage );
			cvResetImageROI( pDebugImage );
		}

		if ( bRun ){
			//Converto l'immagine originale in hsv
			cvCvtColor( pImgOriginal, pImgHsv, CV_RGB2HSV );

			if ( bFirstRun ){
				cvZero( pMask );

				//filtro per valori bassi di saturazoine e filtro per il colore che mi interessa
				//Selezionare col mouse il marker non  accettabile
				cvInRangeS( pImgHsv,
							cvScalar(0 /*MAX( hueValue - 10, 0 )*/, saturationMin, 10, 0 ),
							cvScalar(180 /*MIN( hueValue + 10, 180 )*/, saturationMax, 255, 0 ),
							pMask );


				//Separo il canale hue
				cvSplit( pImgHsv, pHueChannel, NULL, NULL, NULL );

				cvSetImageROI( pHueChannel, sel );
				cvSetImageROI( pMask, sel );
				
				cvCalcHist( &pHueChannel, pHist, 0, pMask );				//calcolo istogramma
				cvGetMinMaxHistValue( pHist, 0, &maxValue, 0, &maxIndex );	//valori massimi
				cvConvertScale( pHist->bins, pHist->bins, maxValue ? 255.0 / maxValue : 0.0, 0 );	//riscalatura

				maxColor = hsv2rgb( maxIndex*180.0 / hueDim );

				cvResetImageROI( pHueChannel );
				cvResetImageROI( pMask );

				search_area = sel;
				bFirstRun = false;

				// DISEGNA L'ISTOGRAMMA
				cvZero( pImgHist );

				int binWidth = pImgHist->width / hueDim;

				for ( int i = 0; i < hueDim; i++ ){
					float binHeight = cvQueryHistValue_1D( pHist, i ) * pImgHist->height / 255;
					CvScalar color = hsv2rgb( i * 180.0 / hueDim );

					cvRectangle( pImgHist,
								 cvPoint( i * binWidth, pImgHist->height ),								//BottomLeft
								 cvPoint( ( i + 1 ) * binWidth, pImgHist->height - binHeight ),			//TopRight
								 color, -1, 8, 0 );
				}
			} else {
				cvInRangeS( pImgHsv,
							cvScalar(0, saturationMin, 10, 0 ),			//HUE, Saturation, Value, Alpha NOT USED
							cvScalar(180, saturationMax, 255, 0 ),
							pMask );

				cvSplit( pImgHsv, pHueChannel, NULL, NULL, NULL );
			}



			cvCalcBackProject( &pHueChannel, pBackProject, pHist );		//backproject
			cvAnd( pBackProject, pMask, pBackProject );

			cvCvtColor( pBackProject, pBackImage, CV_GRAY2BGR );		//lo visualizzo ---> debug

			if ( featCenter.x > 0 && featCenter.y > 0 ){
				cvSet( featMask, CV_RGB(64,64,64) );
				cvRectangle( featMask, cvPoint(featRect.x, featRect.y), cvPoint(featRect.x+featRect.width,featRect.y+featRect.height), CV_RGB(255,255,255),-1);
				cvAnd( pBackProject, featMask, pBackProject );
			}

		/*	CBlobResult *blobs = new CBlobResult( pImgOriginal, 0, 255 ); // test

			for ( int i = 0; i < blobs->GetNumBlobs(); i++ ){
				CBlob *blob = blobs->GetBlob(i);
				blob->FillBlob( pDebugImage, CV_RGB( 255, 0, 0 ) );
			}

			delete blobs;*/

			cvCamShift( pBackProject, search_area,	cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 1, 1 ), &track_comp, &track_box );

			search_area = track_comp.rect;

			track_box.angle = -track_box.angle;

			if ( track_comp.area < 10.0 ){
		//		search_area = cvRect( 0, 0, pBackProject->width, pBackProject->height );
				cout << "*******************************************************************" << endl;
			}


			cout << "Area: " << track_comp.area << endl;
			
			cvEllipseBox( pDebugImage, track_box, CV_RGB(255,0,0), 3, 1, 0 );
			cvRectangle( pDebugImage, cvPoint( search_area.x, search_area.y ), cvPoint( search_area.x + search_area.width, search_area.y + search_area.height ), CV_RGB( 0, 255, 0 ) );

			FlowFeatures( frame );
			TemplateMatch( frame );
		}


		cvShowImage( "original", pDebugImage );
		cvShowImage( "backproject", pBackImage );
		cvShowImage( "hist", pImgHist );

		int c = cvWaitKey(1);

		if ( c == 27 )
			break;
		
		switch ( c ){
			case 'f':
				bInitFeatures = true;
				break;
			case 't':
				bInitTemplate = true;
				break;
		}

	} while ( true );

	cvDestroyAllWindows();

	cvReleaseImage( &pImgOriginal );
	cvReleaseImage( &pImgHsv );
	cvReleaseImage( &pDebugImage );
	cvReleaseImage( &pMask );
	cvReleaseImage( &pBackProject );
	cvReleaseImage( &pBackImage );
	cvReleaseImage( &pImgHist );
	cvReleaseHist( &pHist );


	cvReleaseImage( &pPyramidF );
	cvReleaseImage( &pPrevGrayF );
	cvReleaseImage( &pGrayF );
	cvReleaseImage( &pPrevPyramidF );
	cvReleaseImage( &pEigF );
	cvReleaseImage( &pTempF );

	cvFree( &pPointsF[0] );
	cvFree( &pPointsF[1] );
	cvFree( &statusF );

	cvReleaseCapture(&cam);
}

void FlowFeatures( IplImage *pFrame ){
	if ( !pGrayF ){
		pGrayF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_8U, 1 );
		pPrevGrayF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_8U, 1 );

		pPyramidF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_8U, 1 );
		pPrevPyramidF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_8U, 1 );

		pEigF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_32F, 1 );
		pTempF = cvCreateImage( cvGetSize( pFrame ), IPL_DEPTH_32F, 1 );

		pPointsF[0] = (CvPoint2D32f*)cvAlloc( FEATURE_COUNT * sizeof( pPointsF[0][0] ) );
		pPointsF[1] = (CvPoint2D32f*)cvAlloc( FEATURE_COUNT * sizeof( pPointsF[0][0] ) );
		
		statusF = (char*)cvAlloc( FEATURE_COUNT * sizeof( char ) );
	}

	cvCvtColor( pImgOriginal, pGrayF, CV_RGB2GRAY );

	if ( bInitFeatures ){
		cvSetImageROI( pGrayF, search_area );
		cvSetImageROI( pEigF, search_area );
		cvSetImageROI( pTempF, search_area );

		cvGoodFeaturesToTrack( pGrayF, pEigF, pTempF, pPointsF[1], &feat_count, QUALITY, MIN_DISTANCE, NULL, 3, 1, 0.04 );
		bInitFeatures = false;

		cvResetImageROI( pGrayF );
		cvResetImageROI( pEigF );
		cvResetImageROI( pTempF );

		for ( int i = 0; i < feat_count; i++ ){
			pPointsF[1][i].x += search_area.x;
			pPointsF[1][i].y += search_area.y;
		}

	} else if ( feat_count > 0 ) {
		::cvCalcOpticalFlowPyrLK( pPrevGrayF, pGrayF, pPrevPyramidF, pPyramidF, pPointsF[0], pPointsF[1], feat_count,
								  cvSize(10,10), 3, statusF, 0, cvTermCriteria( CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,10,1), flagsF );
		
		flagsF |= CV_LKFLOW_PYR_A_READY;

		featCenter.x = featCenter.y = featRect.height = featRect.width = 0;
		featRect.x = featRect.y = 1000;
		
		int featCount = 0;

		for ( int i = 0, k = 0; i < feat_count; i++ ){
			if ( !statusF[i] )
				continue;

			if ( pPointsF[1][i].x < search_area.x || pPointsF[1][i].x > search_area.x + search_area.width  ||
				 pPointsF[1][i].y < search_area.y || pPointsF[1][i].y > search_area.y + search_area.height )
				 continue;

			featCenter.x += pPointsF[1][i].x;
			featCenter.y += pPointsF[1][i].y;

			if ( featRect.x > pPointsF[1][i].x )
				featRect.x = pPointsF[1][i].x;

			if ( featRect.y > pPointsF[1][i].y )
				featRect.y = pPointsF[1][i].y;

			pPointsF[1][k++] = pPointsF[1][i];

			cvCircle( pDebugImage, cvPointFrom32f(pPointsF[1][i]), 3, maxColor, -1, 8,0);

			featCount = k;
		}

		for ( int i = 0; i < featCount; i++ ){
			if ( featRect.x + featRect.width < pPointsF[1][i].x )
				featRect.width = pPointsF[1][i].x - featRect.x;

			if ( featRect.y + featRect.height < pPointsF[1][i].y )
				featRect.height = pPointsF[1][i].y - featRect.y;
		}

		if ( featCount > 0 ){
			featCenter.x /= featCount;
			featCenter.y /= featCount;

			cvRectangle( pDebugImage, cvPoint( featRect.x, featRect.y ), cvPoint( featRect.x+featRect.width,featRect.y+featRect.height), maxColor, 1, 8, 0);
			cvCircle( pDebugImage, cvPoint( featRect.x + featRect.width / 2, featRect.y + featRect.height / 2 ), 3, maxColor, 1 ); 
		} else if ( featCount < 2 )
			bInitFeatures = true;
		
	}

	CV_SWAP( pPrevGrayF, pGrayF, pSwapTempF );
	CV_SWAP( pPrevPyramidF, pPyramidF, pSwapTempF );
	CV_SWAP( pPointsF[0], pPointsF[1], pSwapPointsF );
}


void TemplateMatch( IplImage *frame ){
	CvPoint	locMin,
			locMax;

	double valMin,
		   valMax;

	if ( !pImageGrayT ){
		pImageGrayT = cvCreateImage( cvGetSize(frame), IPL_DEPTH_8U, 1 );

		pTemplateT = cvCreateImage( cvGetSize(frame), IPL_DEPTH_8U, 1 );
	}

	cvCvtColor( pImgOriginal, pImageGrayT, CV_RGB2GRAY );

	if ( !bSel && search_area.width > 0 && search_area.height > 0 ){
		if ( bInitTemplate ){
			cvSetImageROI( pImageGrayT, search_area );

			if ( pTemplateT )
				cvReleaseImage( &pTemplateT );

			if ( pMatchResultT )
				cvReleaseImage( &pMatchResultT );

			pTemplateT = cvCreateImage( cvSize( search_area.width, search_area.height ), IPL_DEPTH_8U, 1 );
			pMatchResultT = cvCreateImage( cvSize( pImageGrayT->width - search_area.width + 1, pImageGrayT->height - search_area.height + 1 ), IPL_DEPTH_32F, 1 );

			cvCopy( pImageGrayT, pTemplateT );
			cvResetImageROI( pImageGrayT );

			bInitTemplate = false;
		}

		cvMatchTemplate( pImageGrayT, pTemplateT, pMatchResultT, CV_TM_SQDIFF_NORMED );
		cvMinMaxLoc( pMatchResultT, &valMin, &valMax, &locMin, &locMax, 0 );

		cvRectangle( pDebugImage, locMin, cvPoint( locMin.x + pTemplateT->width, locMin.y + pTemplateT->height ), CV_RGB(255,255,0) );
		
	}

//	cvReleaseImage( &pTemplGrayT );
//	cvReleaseImage( &pImageGrayT );
}

CvScalar hsv2rgb( float hue ){
	int rgb[3], p, sector;
	static const int sector_data[][3] = {{0,2,1}, {1,2,0}, {1,0,2}, {2,0,1}, {2,1,0}, {0,1,2}};
	hue *= 0.033333333333333333333333333333333f;
	sector = cvFloor(hue);
	p = cvRound(255*(hue - sector));
	p ^= sector & 1 ? 255 : 0;

	rgb[sector_data[sector][0]] = 255;
	rgb[sector_data[sector][1]] = 0;
	rgb[sector_data[sector][2]] = p;
	
	return cvScalar(rgb[0], rgb[1], rgb[2], 0 );
}
